import functools
from typing import Callable, List, Optional

from .session import get_current_session
from .wrapped_function import WrappedFunction

__all__ = ["metadata", "collect"]


def metadata(function):
    """Do something."""
    session = get_current_session()
    if session.metadata is not None:
        raise RuntimeError
    session.metadata = function()
    return function


def collect(
    function: Optional[Callable] = None,
    *,
    labels: Optional[List[str]] = None,
    secrets: Optional[List[str]] = None,
):
    """Do something.

    Arguments:
        labels:
        secrets:

    Examples:
        >>> TK
    """

    def wrap(function):
        session = get_current_session()
        wrapped_function = WrappedFunction(function, labels, secrets)
        session.register(wrapped_function)
        return functools.wraps(function)(wrapped_function)

    # Allows for both `@usage.collect` and `@usage.collect(...)`.
    if function:
        return wrap(function)

    return wrap



from __future__ import annotations

from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel


class Error(BaseModel):
    """Schema for an error.

    An ``Error`` object corresponds to an error made from a user call.
    The traceback is pruned such that user function calls are omitted.
    Additionally, secrets are also pruned from the traceback message.

    Attributes:
        line: The line on which the error was raised.
        module: The name of the module in which the error was raised.
        traceback: The traceback message of the error.
    """

    traceback: str
    line: int
    module: str
    type: Type[Any]


class Call(BaseModel):
    """Model for a function call.

    Attributes:
        error: An error, if an error has been raised in this function call.
    """

    error: Optional[Error] = None


class Function(BaseModel):
    """Model for a function.

    Each ``Function`` object corresponds to a method in the developer source code being
    collected by our framework.

    Attributes:
        name: The qualified name of the function.
        labels: A list of labels to categorize the function being tracked.
        calls: A list of calls to this function.
    """

    name: str
    labels: Optional[List[str]] = None
    calls: Optional[List[Call]] = None


class Session(BaseModel):
    """Model for a session.

    A ``Session`` object contains the collected function call history from a
    single program run.

    Attributes:
        uid: A string unique identifier of the user.
        functions: A dictionary of the various tracked functions that were
                   called.
        metadata: A dictionary of global metadata for the session.
    """

    uid: str
    functions: List[Function]
    metadata: Optional[Dict[str, str]] = None



from __future__ import annotations

import atexit
from typing import TYPE_CHECKING, Dict, Optional, Set

from . import client, models

if TYPE_CHECKING:
    from .wrapped_function import WrappedFunction

__all__ = ["init", "shutdown"]

_session = None  # pylint: disable=invalid-name


def is_initialized() -> bool:
    """Do something."""
    return _session is not None


def init(session: Optional[Session] = None) -> None:
    """Initialize PyUsage.

    Arguments:
        session: .
    """
    global _session  # pylint: disable=invalid-name, global-statement
    if session is None:
        _session = Session()
    else:
        _session = session


def shutdown() -> None:
    """Do something."""
    global _session  # pylint: disable=invalid-name, global-statement
    if not is_initialized():
        return

    client.send(_session.model)
    _session = None


atexit.register(shutdown)


def get_current_session() -> Session:
    """Do something."""
    if not is_initialized():
        init()

    assert _session is not None
    return _session


class Session:
    """Is something."""

    def __init__(self):
        """Do something."""
        self._metadata = None
        self._functions = set()

    @property
    def metadata(self) -> Optional[Dict[str, str]]:
        """Do something."""
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: Dict[str, str]) -> None:
        """Do something."""
        assert self._metadata is None
        self._metadata = metadata

    @property
    def functions(self) -> Set[WrappedFunction]:
        """Do something."""
        return self._functions

    @property
    def model(self) -> models.Session:
        """Do something."""
        return models.Session(
            uid="adsf",
            metadata=self.metadata,
            functions=list(function.model for function in self._functions),
        )

    def register(self, function: WrappedFunction) -> None:
        """Do something."""
        self._functions.add(function)



import sys
import traceback
from typing import Callable, List, Optional

from .models import Call, Error, Function


def get_qualified_name(func):
    """Returns the fully qualified name of a function."""
    return f"{func.__module__}.{func.__name__}"


class WrappedFunction:
    """A something."""

    def __init__(
        self,
        function: Callable,
        labels: Optional[List[str]] = None,
        secrets: Optional[List[str]] = None,
    ):
        """Do something."""
        self._labels = labels if labels is not None else []
        self._secrets = secrets if secrets is not None else []
        self._name = get_qualified_name(function)
        self._calls: List[Call] = []

        self._function = function

    def __call__(self, *args, **kwargs):
        """Do something."""
        error = None
        try:
            output = self._function(*args, **kwargs)
            self.calls.append(Call())
            return output
        except Exception as exception:
            error_type, _, error_tb = sys.exc_info()
            filename = error_tb.tb_frame.f_code.co_filename
            line_number = error_tb.tb_lineno
            traceback_msg = traceback.format_exc()
            error = Error(
                traceback=traceback_msg,
                line=line_number,
                module=filename,
                type=error_type,
            )
            self.calls.append(Call(error=error))
            raise exception

    @property
    def labels(self) -> List[str]:
        """Do something."""
        return self._labels

    @property
    def secrets(self) -> List[str]:
        """Do something."""
        return self._secrets

    @property
    def name(self) -> str:
        """Do something."""
        return self._name

    @property
    def calls(self) -> List[Call]:
        """Do something."""
        return self._calls

    @property
    def model(self) -> Function:
        """Do something."""
        return Function(name=self.name, labels=self.labels, calls=self.calls)
